/* d_mos2.cc
 *$Header: /al/acs/src/RCS/d_mos2.cc,v 8.3 95/04/28 20:29:32 al Exp $
 * mos model equations: spice level 2 equivalent
 */
#include "ecah.h"
#include "d_mos.h"
#include "error.h"
#include "io.h"
#include "s__.h"
#include "u_option.h"
#include "declare.h"
/*--------------------------------------------------------------------------*/
//	void	DEV_MOS::eval_mos2(CARD*);
/*--------------------------------------------------------------------------*/
#define short_channel	(m->xj != NOT_INPUT  &&  m->xj > 0.)
#define do_subthreshold	(m->nfs != 0.)
/*--------------------------------------------------------------------------*/
/*static*/ void DEV_MOS::eval_mos2(CARD *brh)
{
  DEV_MOS* This = (DEV_MOS*)brh;
  assert(This);
  const MOS_COMMON* cc = (const MOS_COMMON*)This->common;
  assert(cc);
  const MODEL_MOS* m = (const MODEL_MOS*)cc->model;
  assert(m);

  double sarg, dsarg_dvbs/*ds*/;
  double bargx;
  double gamma_s/*mb*/, dgamma_s_dvds/*ds*/, dgamma_s_dvbs/*mb*/;
  double beta;
  double body;
  double dvdsat_dvgs = 0./*m*/;
  double dvdsat_dvbs = 0./*mb*/;
  double vgsx;
  double vdsx;
  double ufact;
  double dudvgs/*m*/, dudvds/*ds*/, dudvbs/*mb*/;
  double clfact;
  double dldvgs/*m*/, dldvds/*ds*/, dldvbs/*mb*/;
  double dodvbs/*mb*/;
  double xn = 0.;
  double vtxn = 0.;
  double expg;
  double dxndvb =0./*mb*/;	/* subthreshold only */
  double ids_on, didvds;
  double vt/*ds*/;
  double v_phi_d, v_phi_s, v_phi_ds;
  double d2sdb2;
  double barg, dbarg_dvbs, d2bdb2;
  double bsarg, dbsarg_dvbs;
  double vbdsat;
  double dgddb2;
  double vc, vbin, vc_eta;
  double bodys;
  double lambda;
  double vbd, vbs;
  double ueff = 0.;

  int use_vmax;

  use_vmax = m->vmax != NOT_INPUT;
  vt = (K/Q) * SIM::temp;
  
  vbs = This->vbs;
  v_phi_s = m->phi - vbs;
  if (vbs <= 0.){
    sarg = sqrt(v_phi_s);
    dsarg_dvbs = -.5 / sarg;
    d2sdb2 = .5 * dsarg_dvbs / v_phi_s;
    This->sbfwd = NO;
  }else{
    double sphi3 = pow(m->phi, 1.5);
    if (OPT::mosflags & 0001  &&  vbs > m->phi){
      error(bDANGER,"%s: vbs(%g) > phi(%g)\n", brh->printlabel(),vbs,m->phi);
      vbs = m->phi;
      v_phi_s = 0.;
    }
    sarg = sqrt(m->phi) / (1. + .5 * vbs / m->phi);
    dsarg_dvbs = -.5 * sarg * sarg / sphi3;
    d2sdb2 = -dsarg_dvbs * sarg / sphi3;
    This->sbfwd = YES;
    if (!IO::suppresserrors){
      error((This->vbs > m->phi) ? bPICKY : bTRACE,
	    "%s: source fwd biased. vbs=%g\n", brh->printlabel(), This->vbs);
    }
  }
  
  if (OPT::mosflags & 0004){
    vbd = vbs - This->vds;
  }else{
    vbd = This->vbs - This->vds;
  }
  v_phi_d = m->phi - vbd;
  if (vbd <= 0.){
    barg = sqrt(v_phi_d);
    dbarg_dvbs = -.5 / barg;
    d2bdb2 = .5 * dbarg_dvbs / v_phi_d;
    This->dbfwd = NO;
  }else{
    double sphi3 = pow(m->phi, 1.5);
    if (OPT::mosflags & 0002  &&  vbd > m->phi){
      error(bDANGER,"%s: vbd(%g) > phi(%g)\n", brh->printlabel(),vbd,m->phi);
      vbd = m->phi;
      v_phi_d = 0.;
    }
    barg = sqrt(m->phi) / (1. + .5 * vbd / m->phi);
    dbarg_dvbs = -.5 * barg * barg / sphi3;
    d2bdb2 = -dbarg_dvbs * barg / sphi3;
    This->dbfwd = YES;
    if (!IO::suppresserrors){
      error((vbd > m->phi) ? bPICKY : bTRACE,
		"%s: drain fwd biased. vbd=%g\n",
		brh->printlabel(),  This->vbs - This->vds);
    }
  }
  
  if (short_channel){
    double wd = m->xd * barg;
    double argxd = 1. + 2.*wd/m->xj;
    double argd = sqrt(argxd);
    double alpha_d = cc->relxj * (argd - 1.);
    double dalpha_d_dvds = m->xd / (4. * cc->le * argd * barg);
    double dalpha_d_dvbs = -dalpha_d_dvds;
    
    double ws = m->xd * sarg;
    double argxs = 1. + 2.*ws/m->xj;
    double args = sqrt(argxs);
    double alpha_s = cc->relxj * (args - 1.);
    double dalpha_s_dvbs = -m->xd / (4. * cc->le * args * sarg);
    
    gamma_s = m->gamma * (1. - alpha_s - alpha_d);
    dgamma_s_dvds = -m->gamma *  dalpha_d_dvds;
    dgamma_s_dvbs = -m->gamma * (dalpha_d_dvbs + dalpha_s_dvbs);

    double dasdb2 = -m->xd*(d2sdb2 + dsarg_dvbs*dsarg_dvbs*m->xd/(m->xj*argxs))
	/ (cc->le*args);
    double daddb2 = -m->xd*(d2bdb2 + dbarg_dvbs*dbarg_dvbs*m->xd/(m->xj*argxd))
	/ (cc->le*argd);
    dgddb2 = -.5 * m->gamma * (dasdb2 + daddb2);

    if (gamma_s <= 0. && m->gamma > 0. && !IO::suppresserrors){
      error(bWARNING, "%s: gamma is negative\n", brh->printlabel());
      error(bTRACE  , "+   gamma_s=%g, alpha_s=%g, alpha_d=%g\n",
	    gamma_s,    alpha_s,    alpha_d);
    }
  }else{
    gamma_s = m->gamma;
    dgamma_s_dvds = dgamma_s_dvbs = 0.;
    dgddb2 = 0.;
  }
  
  
  vbin = m->vbi + cc->eta_1 * v_phi_s;
  This->von = vbin + gamma_s * sarg;
  
  dodvbs = -cc->eta_1 + dgamma_s_dvbs * sarg + gamma_s * dsarg_dvbs;
  if (do_subthreshold){
    double cdonco = -(gamma_s * dsarg_dvbs + dgamma_s_dvbs * sarg) + cc->eta_1;
    dxndvb = 2. * dgamma_s_dvbs * dsarg_dvbs
      + gamma_s * d2sdb2 + dgddb2 * sarg;
    dodvbs += vt * dxndvb;
    xn = 1. + m->cfsox + cdonco;
    vtxn = vt * xn;
    This->von += vtxn;
    This->subthreshold = (This->vgs < This->von);
    This->cutoff = NO;
  }else if (This->vgs < This->von){
    This->cutoff = YES;
    This->subthreshold = NO;
    This->ids = 0.;
    This->gm = 0.;
    if (OPT::mosflags == 0100){
      This->gds = OPT::gmin;
    }else{
      This->gds = 0.;
    }
    This->gmb = 0.;
    This->vgst = This->vgs - This->von;
    return;
  }else{
    This->cutoff = This->subthreshold = NO;
  }

  vgsx = (This->subthreshold) ? This->von : This->vgs;
  vc = vgsx - vbin;
  vc_eta = vc / cc->eta;
  This->vgst = This->vgs - This->von;

  if (m->uexp != NOT_INPUT  &&  This->vgst > m->vbp){
    ufact = pow(m->vbp/This->vgst, m->uexp);
    dudvgs = -ufact * m->uexp / This->vgst;
    dudvds = 0.;	/* wrong, but as per spice2 */
    dudvbs = dodvbs * ufact * m->uexp / This->vgst;
  }else{
    ufact = 1.;
    dudvgs = dudvds = dudvbs = 0.;
  }

  if (use_vmax){
    double y3;
    double xvalid = 0.;
    double x4[8];
    double sarg3 = sarg*sarg*sarg;
    double gammad = gamma_s / cc->eta;
    ueff = m->uo * ufact;	/* ???? */
    double v1 = vc_eta + v_phi_s;
    double v2 = v_phi_s;
    double xv = m->vmax * cc->le / ueff;
    double a1 = gammad * (4./3.);
    double b1 = -2. * (v1+xv);
    double c1 = -2. * gammad * xv;			/* end of scope */
    double d1 = 2.*v1*(v2+xv) - v2*v2 - (4./3.)*gammad*sarg3;
    double a = -b1;					/* xv, v1, v2, sarg3 */
    double b = a1 * c1 - 4. * d1;
    double c = -d1 * (a1*a1 - 4.*b1) - c1*c1;
    double r = -a*a / 3. + b;
    double s = 2. * a*a*a / 27. - a*b / 3. + c;		/* b, c */
    double r3 = r*r*r;					/* r */
    double s2 = s*s;
    double p = s2 / 4. + r3 / 27.;			/* r3 */
    double p0 = fabs(p);
    double p2 = sqrt(p0);
    if (p < 0.){					/* p */
      double ro = pow((s2 / 4. + p0), (1./6.));		/* s2, p0 */
      double fi = atan(-2. * p2 / s);
      y3 = 2. * ro * cos(fi/3.) - a / 3.;
    }else{
      double p3 = pow((fabs(-s/2.+p2)), (1./3.));
      double p4 = pow((fabs(-s/2.-p2)), (1./3.));	/* s, p2 */
      y3 = p3 + p4 - a / 3.;				/* a */
    }

    int iknt = 0;
    if (a1*a1 / 4. - b1 + y3  < 0.  &&  y3*y3 / 4. - d1  < 0.){
      error(bWARNING,
	    "%s: internal error: a3,b4, a1=%g, b1=%g, y3=%g, d1=%g\n",
	    brh->printlabel(),	      a1,    b1,    y3,    d1);
    }else{
      double a3 = sqrt(a1*a1 / 4. - b1 + y3);
      double b3 = sqrt(y3*y3 / 4. - d1);
      for (int i = 0;   i < 4;   i++){
	static const double sig1[4] = {1., -1., 1., -1.};
	static const double sig2[4] = {1., 1., -1., -1.};
	double a4 = a1 / 2. + sig1[i] * a3;
	double b4 = y3 / 2. + sig2[i] * b3;		/* y3 */
	double delta4 = a4*a4 / 4. - b4;
	if (delta4 >= 0.){
	  x4[iknt++] = - a4 / 2. + sqrt(delta4);
	  x4[iknt++] = - a4 / 2. - sqrt(delta4);	/* i */
	}
      }
    }

    int root_count = 0;
    for (int j = 0;   j < iknt;   j++){			/* iknt */
      if (x4[j] > 0.){
	double poly4 = x4[j]*x4[j]*x4[j]*x4[j]	/* ~= 0, used as check	*/
	  	+ a1 * x4[j]*x4[j]*x4[j]	/* roundoff error not	*/
	 	+ b1 * x4[j]*x4[j]		/* propagated, so ok	*/
		+ c1 * x4[j]
		+ d1;				/* a1, b1, c1, d1 */
	if (fabs(poly4) <= 1e-6){
	  root_count++;
	  if (root_count <= 1)			/* xvalid = min(x4[j]) */
	    xvalid=x4[j];
	  if (x4[j] <= xvalid)
	    xvalid=x4[j];				/* x4[], j */
	}
      }
    }
    if (root_count <= 0){				/* root_count */
      if (!IO::suppresserrors)
	error(bWARNING, "%s: Baum's theory rejected\n", brh->printlabel());
      use_vmax = NO;
    }else{
      This->vdsat = xvalid*xvalid - v_phi_s;
    }
  }
  
  if (!use_vmax){
    if (gamma_s > 0.){
      double argv = vc_eta + v_phi_s;
      if (argv > 0.){
	double gammad = gamma_s / cc->eta;
	double gammd2 = gammad * gammad;
	double arg = sqrt(1. + 4. * argv / gammd2);
	This->vdsat = vc_eta  +  gammd2 * (1.-arg) / 2.;
	dvdsat_dvgs = (1. - 1./arg) / cc->eta;
	dvdsat_dvbs = (gammad * (1.-arg) + 2.*argv / (gammad*arg))
	  / cc->eta * dgamma_s_dvbs
	    + 1./arg + cc->eta_1 * dvdsat_dvgs;
      }else{
	This->vdsat = 0.;
	dvdsat_dvgs = dvdsat_dvbs = 0.;
	if (!IO::suppresserrors){
	  error(bWARNING, "%s: argv is negative\n", brh->printlabel());
	  error(bTRACE  , "+   vc=%g, argv=%g, vdsat=%g\n",
		vc,    argv,    This->vdsat);
	  error(bTRACE  , "+   vds=%g, vgs=%g, vbs=%g\n",
		This->vds, This->vgs, This->vbs);
	}
      }
    }else{
      This->vdsat = vc_eta;
      dvdsat_dvgs = 1.;
      dvdsat_dvbs = 0.;
    }
  }
  
  if (This->vdsat < 0.){
    error(bWARNING, "%s: calculated vdsat (%g) < 0.  using vdsat = 0.\n",
	  brh->printlabel(), This->vdsat);
    This->vdsat = 0.;
  }
  
  vbdsat = vbs - This->vdsat;
  v_phi_ds = m->phi - vbdsat;
  if (vbdsat <= 0.){
    bsarg = sqrt(v_phi_ds);
    dbsarg_dvbs = -.5 / bsarg;
  }else{
    double sphi3 = pow(m->phi, 1.5);
    bsarg = sqrt(m->phi) / (1. + .5 * vbdsat / m->phi);
    dbsarg_dvbs = -.5 * bsarg * bsarg / sphi3;
  }
  bodys = bsarg*bsarg*bsarg - sarg*sarg*sarg;
  
  if (use_vmax){
    double gdbdvs =
      2. * gamma_s * (bsarg*bsarg*dbsarg_dvbs - sarg*sarg*dsarg_dvbs);
    double argv = vc_eta - This->vdsat;
    double vqchan = argv - gamma_s * bsarg;
    double dqdsat = -1. + gamma_s * dbsarg_dvbs;
    double vl = m->vmax * cc->le;
    double dfunds = vl * dqdsat - ueff * vqchan;
    double dfundg = (vl - ueff * This->vdsat) / cc->eta;
    double dfundb = -vl * (1. + dqdsat - cc->eta_1 / cc->eta)
      + ueff * (gdbdvs - dgamma_s_dvbs * bodys / 1.5) / cc->eta;
    dvdsat_dvgs = -dfundg / dfunds;
    dvdsat_dvbs = -dfundb / dfunds;
  }
  
  if (m->lambda == NOT_INPUT){
    if (This->vds != 0.){
      double dldsat;
      if (use_vmax){
	double xdv = m->xd / sqrt(m->neff);
	double xlv = m->vmax * xdv / (2. * ueff);
	double argv = This->vds - This->vdsat;
	if (argv < 0.)
	  argv = 0.;
	double xls = sqrt(xlv*xlv + argv);
	lambda = (xls-xlv) * xdv / (cc->le * This->vds);
	dldsat = xdv / (2. * xls * cc->le);
      }else{
	double argv = (This->vds - This->vdsat) / 4.;
	double sargv = sqrt(1. + argv*argv);
	if (argv + sargv >= 0.){
	  double dl = m->xd * sqrt(argv + sargv);
	  lambda = dl / (cc->le * This->vds);
	  dldsat = lambda * This->vds / (8. * sargv);
	}else{
	  lambda = 0.;
	  dldsat = 0.;
	  error(bWARNING, "%s: internal error: vds(%g) < vdsat(%g)\n",
		brh->printlabel(), This->vds,   This->vdsat);
	}
      }
      dldvgs =  dvdsat_dvgs * dldsat;
      dldvds =              - dldsat;
      dldvbs =  dvdsat_dvbs * dldsat;
    }else{
      lambda = 0.;
      dldvgs = dldvds = dldvbs = 0.;
    }
  }else{
    lambda = m->lambda;
    dldvgs = dldvbs = 0.;
    dldvds = -lambda;
  }
  
  clfact = (1. - lambda * This->vds);
  if (clfact < m->xwb/cc->le){
    double leff = m->xwb / (2. - (clfact * cc->le / m->xwb));
    double dfact = (leff * leff) / (m->xwb * m->xwb);
    clfact = leff / cc->le;
    dldvgs *= dfact;
    dldvds *= dfact;
    dldvbs *= dfact;
  }
  
  This->saturated = (This->vds > This->vdsat);
  vdsx =  (This->saturated) ? This->vdsat : This->vds;
  bargx = (This->saturated) ? bsarg : barg;
  body = bargx*bargx*bargx - sarg*sarg*sarg;
  expg = (This->subthreshold) ? exp(This->vgst / vtxn) : 1.;
  beta = cc->beta * ufact / clfact;
  
  ids_on = beta * ((vc - cc->eta_2 * vdsx) * vdsx  - (2./3.) * gamma_s * body);
  didvds = beta * (vc  -  cc->eta * vdsx  -  gamma_s * bargx);
  
  This->ids = ids_on * expg;
  
  This->gm = beta * vdsx;
  This->gm += ids_on * (dudvgs/ufact - dldvgs/clfact);
  if (This->saturated)
    This->gm += didvds * dvdsat_dvgs;
  if (This->subthreshold){
    This->gm = ids_on / vtxn;
    if (This->saturated)
      This->gm += didvds * dvdsat_dvgs;
    This->gm *= expg;
  }
  This->gds = (This->saturated) ? 0.: didvds;
  This->gds += ids_on * (dudvds/ufact - dldvds/clfact);
  if (short_channel)
    This->gds -= beta * (2./3.) * body * dgamma_s_dvds;
  if (This->subthreshold){
    double dxndvd = dgamma_s_dvds * dsarg_dvbs;
    double dodvds = dgamma_s_dvds * sarg + vt * dxndvd;
    double gmw = This->ids * This->vgst / (vtxn * xn);
    This->gds *= expg;
    This->gds -= This->gm * dodvds + gmw * dxndvd;
  }
  This->gmb = beta * (cc->eta_1 * vdsx - gamma_s * (sarg - bargx));
  This->gmb += ids_on * (dudvbs/ufact - dldvbs/clfact);
  if (short_channel)
    This->gmb -= beta * (2./3.) * body * dgamma_s_dvbs;
  if (This->saturated)
    This->gmb += didvds * dvdsat_dvbs;
  if (This->subthreshold){
    double gmw = This->ids * This->vgst / (vtxn * xn);
    This->gmb += beta * dodvbs * vdsx;
    This->gmb *= expg;
    This->gmb -= This->gm * dodvbs + gmw * dxndvb;
  }
}
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
